/* This code was written by Jeff Dunlop of DB/Soft Publishing Co.,
 * and is hereby released into the public domain.  No copyright
 * can exist on this code, and no restriction may be placed on
 * its use.
 */

/* Note:  db_err() is a function that checks the passed value
 * (a PX* function call return value) for non-zero and decides what to
 * do with it.  All error-trapping strategies are different from
 * programmer to programmer, but this allows me to standardize
 * fatal error trapping.  Db_err() btw is a function that relies on
 * another library, which is the reason it is not included here.  I'm
 * pretty sure that the non-engine specific code included will compile
 * with most/all other compilers, so most/all engine users should be
 * fine.
 */

/* The calling convention for px_sort is as follows:
 * px_handle is the handle of the source table.
 * px_recbuf is the handle to the source transfer buffer
 * field_list is a null termed array of fieldnames.  The fieldnames
 *      MUST be valid, and MUST be null termed.  Departure from this
 *      convention will cause an 'invalid field' error from PX.  The
 *      fields MUST be in the order in which they are to appear in the
 *      new table.
 * num_keyfields is the number of fields you want in your key.  Per
 *      Paradox, a primary key can be any number of fields, but the
 *      fields must start with the first, and continue with consecutive
 *      fields until done.
 * tbl_name is the name of the destination table.  It must include
 *      any path as needed, and should not include an extension.
 */

/* When specifying the field order and number of fields in the key,
 * remember that PX will delete (not include) any records that would
 * cause a key violation.  So, if you don't want any records excluded, be
 * sure to include the original key field(s) in the new key.  If you
 * don't want the original key field(s) to affect the sort, you can
 * always place them after the fields that you DO want to affect the
 * sort.

/*----------------------------[ px_sort.c ]----------------------------*/
/* Routines germaine to building a new table from an old table, in     */
/* any sort order.                                                     */
/*---------------------------------------------------------------------*/

/*-----------------------------[ headers ]-----------------------------*/
#include "pxengine.h"
#include <dos.h>

/*----------------------------[ prototypes ]---------------------------*/
int tbl_build(TABLEHANDLE px_handle, char *field_list[], char *tbl_name);
int px_sort(TABLEHANDLE px_handle, RECORDHANDLE px_recbuf,
             char *field_list[], int num_keyfields, char *tbl_name);

/*---------------------------------------------------------------------*/
/*---------------------------[ functions ]-----------------------------*/
/*---------------------------------------------------------------------*/

/*----------------------------[ tbl_build ]----------------------------*/
/* Build a table identical to source table, with fields in new order.  */
/*---------------------------------------------------------------------*/
/* input:                                                              */
/*      px_handle: handle to source table                              */
/*      field_list: list of fieldnames in order of appearance          */
/*      tbl_name: name of table to create                              */
/* local:                                                              */
/*      type_bfr = buffer for list of field types                      */
/*      type_list = array of pointers into type_bfr                    */
/*      i = loop counter                                               */
/*      px_result = PX function return value                           */
/* return:                                                             */
/*      PXSUCCESS on success, PX error number on error                 */
/*---------------------------------------------------------------------*/
int tbl_build(TABLEHANDLE px_handle, char *field_list[], char *tbl_name)
{
    char type_bfr[255][4];
    char *type_list[255];
    int i = 0,
        px_result;

    for (i = 0; i <= 255; i++)
        type_list[i] = &type_bfr[i][0];

    i = 0;
    while (field_list[i][0] != '\0')
    {
        if ((px_result = PXFldType(px_handle, get_pxfield(px_handle,
            field_list[i]), 5, type_list[i])) != PXSUCCESS)
            return px_result;
        i++;
    }
    return (PXTblCreate(tbl_name, i, field_list,
        type_list));
}

/*-------------------------[ get_pxfield ]----------------------------*/
/*                    Get field number for field                      */
/*--------------------------------------------------------------------*/
/* input:                                                             */
/*      px_handle = table handle                                      */
/*      px_fieldname = field name                                     */
/* local:                                                             */
/*      px_field = field handle                                       */
/* return:                                                            */
/*      field handle, or no return on fatal error                     */
/---------------------------------------------------------------------*/
int get_pxfield(TABLEHANDLE px_handle, char *px_fieldname)
{
    FIELDHANDLE px_field;

    db_err(PXFldHandle(px_handle, px_fieldname, &px_field));
    return px_field;
}


/*----------------------------[ px_sort ]------------------------------*/
/*           Build a sorted table from an existing table               */
/*---------------------------------------------------------------------*/
/* input:                                                              */
/*      px_handle: handle to source table                              */
/*      px_recbuf: handle to source transfer buffer                    */
/*      field_list: list of fieldnames in key order                    */
/*      num_keyfields: number of fields to comprise the key            */
/*      tbl_name: destination table name (path + name, no ext)         */
/* local:                                                              */
/*      i: loop counter                                                */
/*      field_type: buffer for type of current field                   */
/*      new_pxhandle: handle to dest table                             */
/*      new_pxrecbuf: handle to dest transfer buffer                   */
/*      *_temp: generic typed transfer buffer transfer locations       */
/*      cur_recnum = record number of current record                   */
/*      total_recnum = number of records in table                      */
/* return:                                                             */
/*      PXSUCCESS if successful, PXErrNo if not                        */
/*---------------------------------------------------------------------*/
int px_sort(TABLEHANDLE px_handle, RECORDHANDLE px_recbuf,
             char *field_list[], int num_keyfields, char *tbl_name)
{
    int i = 0,
        px_result;
    char field_type[5],
         alpha_temp[256];
    double double_temp;
    short short_temp;
    long cur_recnum = 1,
         total_recnum;
    DATE date_temp;
    TABLEHANDLE new_pxhandle;
    RECORDHANDLE new_pxrecbuf;
    FIELDHANDLE px_fieldnum,
                field_handles[] = {{1}, {2}};

    tbl_build(px_handle, field_list, tbl_name);
    if ((px_result = PXTblOpen(tbl_name, &new_pxhandle, 0, FALSE))
        != PXSUCCESS)
        return px_result;
    if ((px_result = PXRecBufOpen(new_pxhandle, &new_pxrecbuf)) != PXSUCCESS)
        return px_result;
    PXTblNRecs(px_handle, &total_recnum);
    db_err(PXRecFirst(px_handle));
    db_err(PXRecGet(px_handle, px_recbuf));
    while (cur_recnum <= total_recnum)
    {
        i = 0;
        while (field_list[i][0] != '\0')
        {
            if ((px_result = PXFldType(px_handle,
                px_fieldnum = get_pxfield(px_handle, field_list[i]),
                5, field_type)) != PXSUCCESS)
                return px_result;
            switch(field_type[0])
            {
                case 'A':
                {
                    db_err(PXGetAlpha(px_recbuf, px_fieldnum,
                        sizeof(alpha_temp), alpha_temp));
                    db_err(PXPutAlpha(new_pxrecbuf, i + 1, alpha_temp));
                    break;
                }
                case 'N':
                case '$':
                {
                    PXGetDoub(px_recbuf, px_fieldnum, &double_temp);
                    PXPutDoub(new_pxrecbuf, i + 1, double_temp);
                    break;
                }
                case 'S':
                {
                    PXGetShort(px_recbuf, px_fieldnum, &short_temp);
                    PXPutShort(new_pxrecbuf, i + 1, short_temp);
                    break;
                }
                case 'D':
                {
                    PXGetDate(px_recbuf, px_fieldnum, &date_temp);
                    PXPutDate(new_pxrecbuf, i + 1, date_temp);
                    break;
                }
            }
            i++;
        }
        db_err(PXRecAppend(new_pxhandle, new_pxrecbuf));
        if (cur_recnum < total_recnum)
        {
            db_err(PXRecNext(px_handle));
            db_err(PXRecGet(px_handle, px_recbuf));
        }
        cur_recnum++;
    }

    /* sort the new table */
    PXRecBufClose(new_pxrecbuf);
    PXTblClose(new_pxhandle);
    db_err(PXKeyAdd(tbl_name, num_keyfields, field_handles, PRIMARY));
    return PXSUCCESS;
}

